@@ -14,6 +14,7 @@ import openmdao.utils.hooks as hooks
 from openmdao.test_suite.components.paraboloid import Paraboloid
 from openmdao.test_suite.components.sellar import SellarDerivatives
 from openmdao.utils.units import convert_units
+from openmdao.warnings import DerivativesWarning
 
 try:
     from parameterized import parameterized
@@ -855,6 +856,26 @@ class TestProblem(unittest.TestCase):
 
         prob.check_totals(method='cs')
 
+    def test_set_cs_error_messages(self):
+        prob = om.Problem()
+        prob.model.add_subsystem('comp', Paraboloid())
+        prob.setup()
+        prob.run_model()
+        with self.assertRaises(RuntimeError) as cm:
+            prob.set_complex_step_mode(True)
+
+        msg = "Problem: To enable complex step, specify 'force_alloc_complex=True' when calling " + \
+            "setup on the problem, e.g. 'problem.setup(force_alloc_complex=True)'"
+        self.assertEqual(cm.exception.args[0], msg)
+
+        prob = om.Problem()
+        prob.model.add_subsystem('comp', Paraboloid())
+        with self.assertRaises(RuntimeError) as cm:
+            prob.set_complex_step_mode(True)
+        msg = "Problem: set_complex_step_mode cannot be called before `Problem.run_model()`, " + \
+            "`Problem.run_driver()`, or `Problem.final_setup()`."
+        self.assertEqual(cm.exception.args[0], msg)
+
     def test_feature_run_driver(self):
         import numpy as np
 
@@ -1396,7 +1417,7 @@ class TestProblem(unittest.TestCase):
               "You chose 'fwd' for a problem with 99 design variables and 10 " \
               "response variables (objectives and nonlinear constraints)."
 
-        with assert_warning(RuntimeWarning, msg):
+        with assert_warning(DerivativesWarning, msg):
             prob.final_setup()
 
     def test_setup_bad_mode_direction_rev(self):
@@ -1418,7 +1439,7 @@ class TestProblem(unittest.TestCase):
               "You chose 'rev' for a problem with 10 design variables and 20 " \
               "response variables (objectives and nonlinear constraints)."
 
-        with assert_warning(RuntimeWarning, msg):
+        with assert_warning(DerivativesWarning, msg):
             prob.final_setup()
 
     def test_run_before_setup(self):
@@ -2031,6 +2052,55 @@ class TestProblem(unittest.TestCase):
         self.assertRegex(output[12], r'^\s+upper:')
         self.assertRegex(output[13], r'^\s+array+\(+\[[0-9., e+-]+\]+\)')
 
+    def test_list_problem_vars_driver_scaling(self):
+        model = SellarDerivatives()
+        model.nonlinear_solver = om.NonlinearBlockGS()
+
+        prob = om.Problem(model)
+        prob.driver = om.ScipyOptimizeDriver()
+        prob.driver.options['optimizer'] = 'SLSQP'
+        prob.driver.options['tol'] = 1e-9
+
+        model.add_design_var('z', lower=np.array([-10.0, 0.0]),
+                             upper=np.array([10.0, 10.0]), ref=1.5)
+        model.add_design_var('x', lower=0.0, upper=10.0)
+        model.add_objective('obj', ref=3.0)
+        model.add_constraint('con1', upper=0.0)
+        model.add_constraint('con2', upper=0.0, ref=2.0)
+
+        prob.setup()
+        prob.run_driver()
+
+        # Driver Scaling
+        stdout = sys.stdout
+        strout = StringIO()
+        sys.stdout = strout
+
+        try:
+            prob.list_problem_vars()
+        finally:
+            sys.stdout = stdout
+        output = strout.getvalue().split('\n')
+
+        self.assertTrue('1.31' in output[5]) # z
+        self.assertTrue('-10.' in output[14]) # con
+        self.assertTrue('1.06' in output[21]) # obj
+
+        # Model Scaling
+        stdout = sys.stdout
+        strout = StringIO()
+        sys.stdout = strout
+
+        try:
+            prob.list_problem_vars(driver_scaling=False)
+        finally:
+            sys.stdout = stdout
+        output = strout.getvalue().split('\n')
+
+        self.assertTrue('1.9' in output[5]) # z
+        self.assertTrue('-20.' in output[14]) # con
+        self.assertTrue('3.18' in output[21]) # obj
+
     def test_feature_list_problem_vars(self):
         import numpy as np
         import openmdao.api as om
@@ -2101,6 +2171,69 @@ class NestedProblemTestCase(unittest.TestCase):
         p.setup()
         p.run_model()
 
+    def test_cs_across_nested(self):
+
+        class NestedAnalysis(om.ExplicitComponent):
+
+            def __init__(self):
+                super().__init__()
+                self._problem = None
+
+            def setup(self):
+                self.add_input('x', val=0.0)
+                self.add_input('y', val=0.0)
+
+                self.add_output('f_xy', val=0.0)
+
+                # Setup sub-problem
+                self._problem = prob = om.Problem()
+                model = prob.model
+                model.add_subsystem('parab', Paraboloid(), promotes=['*'])
+                prob.setup(force_alloc_complex=True)
+
+            def setup_partials(self):
+                self.declare_partials(of='*', wrt='*')
+
+            def compute(self, inputs, outputs):
+                prob = self._problem
+                under_cs = self.under_complex_step
+
+                if under_cs:
+                    prob.set_complex_step_mode(True)
+
+                # Set inputs
+                prob.set_val('x', inputs['x'])
+                prob.set_val('y', inputs['y'])
+
+                # Run model
+                prob.run_model()
+
+                # Extract outputs
+                outputs['f_xy'] = prob.get_val('f_xy')
+
+                if under_cs:
+                    prob.set_complex_step_mode(False)
+
+            def compute_partials(self, inputs, partials):
+                totals = self._problem.compute_totals(of='f_xy', wrt=['x', 'y'])
+                partials['f_xy', 'x'] = totals['f_xy', 'x']
+                partials['f_xy', 'y'] = totals['f_xy', 'y']
+
+        prob = om.Problem()
+        model = prob.model
+        model.add_subsystem('nested', NestedAnalysis(), promotes=['*'])
+
+        prob.setup(force_alloc_complex=True)
+
+        prob.set_val('x', 3.5)
+        prob.set_val('y', 1.5)
+
+        prob.run_model()
+
+        totals = prob.check_totals(of='f_xy', wrt=['x', 'y'], method='cs', out_stream=None)
+        for key, val in totals.items():
+            assert_near_equal(val['rel error'][0], 0.0, 1e-12)
+
 
 class SystemInTwoProblemsTestCase(unittest.TestCase):
     def test_2problems(self):
