@@ -12,28 +12,24 @@ import time
 import numpy as np
 
 from openmdao.core.constants import INT_DTYPE
-from openmdao.utils.general_utils import ContainsAll, simple_warning, _prom2ivc_src_dict
+from openmdao.utils.general_utils import ContainsAll, _prom2ivc_src_dict
 
-from openmdao.utils.mpi import MPI, multi_proc_exception_check
+from openmdao.utils.mpi import MPI, check_mpi_env, multi_proc_exception_check
 from openmdao.utils.coloring import _initialize_model_approx, Coloring
+from openmdao.warnings import issue_warning, DerivativesWarning
+from openmdao.vectors.vector import _full_slice
 
-# Attempt to import petsc4py.
-# If OPENMDAO_REQUIRE_MPI is set to a recognized positive value, attempt import
-# and raise exception on failure. If set to anything else, no import is attempted.
-if 'OPENMDAO_REQUIRE_MPI' in os.environ:
-    if os.environ['OPENMDAO_REQUIRE_MPI'].lower() in ['always', '1', 'true', 'yes']:
-        from petsc4py import PETSc
-    else:
-        PETSc = None
-# If OPENMDAO_REQUIRE_MPI is unset, attempt to import petsc4py, but continue on failure
-# with a notification.
-else:
+
+use_mpi = check_mpi_env()
+if use_mpi is not False:
     try:
         from petsc4py import PETSc
     except ImportError:
         PETSc = None
-        sys.stdout.write("Unable to import petsc4py. Parallel processing unavailable.\n")
-        sys.stdout.flush()
+        if use_mpi is True:
+            raise ImportError("Importing petsc4py failed and OPENMDAO_USE_MPI is true.")
+elif use_mpi is False:
+    PETSc = None
 
 _contains_all = ContainsAll()
 
@@ -77,7 +73,7 @@ class _TotalJacInfo(object):
     output_list : list of str
         List of names of output variables for this total jacobian.  In fwd mode, outputs
         are responses.  In rev mode, outputs are design variables.
-    output_vec : Dict of vectors keyed by vec_name.
+    output_vec : dict of vectors keyed by vec_name.
         Designated output vectors based on value of fwd.
     owning_ranks : dict
         Map of absolute var name to the MPI process that owns it.
@@ -121,6 +117,8 @@ class _TotalJacInfo(object):
         driver_scaling : bool
             If True (default), scale derivative values by the quantities specified when the desvars
             and responses were added. If False, leave them unscaled.
+        get_remote : bool
+            Whether to get remote variables if using MPI.
         """
         driver = problem.driver
         prom2abs = problem.model._var_allprocs_prom2abs_list['output']
@@ -255,7 +253,7 @@ class _TotalJacInfo(object):
                            "be turned off.\ncoloring design vars: %s, current design vars: "
                            "%s\ncoloring responses: %s, current responses: %s." %
                            (driver_wrt, wrt, driver_of, of))
-                    simple_warning(msg)
+                    issue_warning(msg, category=DerivativesWarning)
                     self.simul_coloring = None
 
             if not isinstance(self.simul_coloring, Coloring):
@@ -804,7 +802,7 @@ class _TotalJacInfo(object):
                 if name in abs2idx and name in slices and name not in self.remote_vois:
                     var_idx = abs2idx[name]
                     slc = slices[name]
-                    if MPI and meta['distributed'] and model.comm.size > 1 and self.get_remote:
+                    if MPI and meta['distributed'] and self.get_remote:
                         if indices is not None:
                             local_idx, sizes_idx, _ = self._dist_driver_vars[name]
 
@@ -944,16 +942,12 @@ class _TotalJacInfo(object):
             Jac setter method.
         """
         coloring = imeta['coloring']
-        both = coloring._fwd and coloring._rev
         input_setter = self.simul_coloring_input_setter
         jac_setter = self.simul_coloring_jac_setter
 
         for color, ilist in enumerate(coloring.color_iter(mode)):
             if len(ilist) == 1:
-                if both:
-                    yield ilist, input_setter, jac_setter, None
-                else:
-                    yield ilist[0], self.single_input_setter, self.single_jac_setter, None
+                yield ilist, input_setter, jac_setter, None
             else:
                 # yield all indices for a color at once
                 yield ilist, input_setter, jac_setter, imeta['itermeta'][color]
@@ -1270,6 +1264,7 @@ class _TotalJacInfo(object):
                 i = loc_idx
             else:
                 return
+
         if mode == 'fwd':
             self.J[jac_idxs[vecname], i] = deriv_val[deriv_idxs[vecname]]
         else:  # rev
@@ -1491,12 +1486,14 @@ class _TotalJacInfo(object):
                             varlist = '(' + ', '.join([name for name in par_print[key]]) + ')'
                             print('Solving color:', key, varlist)
                         else:
-                            print('In mode: %s, Solving variable(s) using simul coloring:' % mode)
                             if key == '@simul_coloring':
+                                print(f'In mode: {mode}, Solving variable(s) using simul '
+                                      'coloring:')
                                 for local_ind in imeta['coloring']._local_indices(inds=inds,
                                                                                   mode=self.mode):
                                     print("   {}".format(local_ind))
                             else:
+                                print('In mode: %s.' % mode)
                                 print_key = key
                                 if key in self.ivc_print_names:
                                     print_key = self.ivc_print_names[key]
@@ -1522,7 +1519,7 @@ class _TotalJacInfo(object):
                                 model._solve_linear(model._lin_vec_names, mode, rel_systems)
 
                     if debug_print:
-                        print('Elapsed Time:', time.time() - t0, '\n', flush=True)
+                        print(f'Elapsed Time: {time.time() - t0} secs\n', flush=True)
 
                     jac_setter(inds, mode)
 
@@ -1534,8 +1531,6 @@ class _TotalJacInfo(object):
             # Debug outputs scaled derivatives.
             self._print_derivatives()
 
-        # np.save("total_jac%d.npy" % self.comm.rank, self.J)
-
         return self.J_final
 
     def compute_totals_approx(self, initialize=False, progress_out_stream=None):
@@ -1562,6 +1557,7 @@ class _TotalJacInfo(object):
         model = self.model
         comm = model.comm
         return_format = self.return_format
+        debug_print = self.debug_print
 
         # Prepare model for calculation by cleaning out the derivatives
         # vectors.
@@ -1605,6 +1601,8 @@ class _TotalJacInfo(object):
         wrt_idx = model._owns_approx_wrt_idx
         wrt_meta = self.wrt_meta
 
+        t0 = time.time()
+
         totals = self.J_dict
         if return_format == 'flat_dict':
             for prom_out, output_name in zip(self.prom_of, of):
@@ -1621,12 +1619,14 @@ class _TotalJacInfo(object):
                         # constraint wrt all other inputs.
                         continue
 
-                    ofidx = of_idx[output_name] if output_name in of_idx else None
-                    wrtidx = wrt_idx[input_name] if input_name in wrt_idx else None
+                    ofidx = of_idx[output_name] if output_name in of_idx else _full_slice
+                    wrtidx = wrt_idx[input_name] if input_name in wrt_idx else _full_slice
 
-                    totals[prom_out, prom_in][:] = _get_subjac(approx_jac[output_name, input_name],
-                                                               prom_out, prom_in, ofidx, wrtidx,
-                                                               dist_resp, comm)
+                    approxJ = approx_jac[output_name, input_name]
+
+                    totals[prom_out, prom_in][:] = _get_approx_subjac(approxJ,
+                                                                      prom_out, prom_in, ofidx,
+                                                                      wrtidx, dist_resp, comm)
 
         elif return_format in ('dict', 'array'):
             for prom_out, output_name in zip(self.prom_of, of):
@@ -1645,22 +1645,26 @@ class _TotalJacInfo(object):
                         # constraint wrt all other inputs.
                         continue
 
-                    ofidx = of_idx[output_name] if output_name in of_idx else None
-                    wrtidx = wrt_idx[input_name] if input_name in wrt_idx else None
+                    ofidx = of_idx[output_name] if output_name in of_idx else _full_slice
+                    wrtidx = wrt_idx[input_name] if input_name in wrt_idx else _full_slice
 
                     if prom_out == prom_in and isinstance(tot[prom_in], dict):
                         rows, cols, data = tot[prom_in]['coo']
-                        data[:] = _get_subjac(approx_jac[output_name, input_name],
-                                              prom_out, prom_in, ofidx, wrtidx,
-                                              dist_resp, comm)[rows, cols]
+                        data[:] = _get_approx_subjac(approx_jac[output_name, input_name],
+                                                     prom_out, prom_in, ofidx, wrtidx,
+                                                     dist_resp, comm)[rows, cols]
                     else:
-                        tot[prom_in][:] = _get_subjac(approx_jac[output_name, input_name],
-                                                      prom_out, prom_in, ofidx, wrtidx,
-                                                      dist_resp, comm)
+                        tot[prom_in][:] = _get_approx_subjac(approx_jac[output_name, input_name],
+                                                             prom_out, prom_in, ofidx, wrtidx,
+                                                             dist_resp, comm)
+
         else:
             msg = "Unsupported return format '%s." % return_format
             raise NotImplementedError(msg)
 
+        if debug_print:
+            print(f'Elapsed time to approx totals: {time.time() - t0} secs\n', flush=True)
+
         # Driver scaling.
         if self.has_scaling:
             self._do_driver_scaling(totals)
@@ -1668,8 +1672,71 @@ class _TotalJacInfo(object):
         if return_format == 'array':
             totals = self.J  # change back to array version
 
+        if debug_print:
+            # Debug outputs scaled derivatives.
+            self._print_derivatives()
+
         return totals
 
+    def check_total_jac(self, raise_error=True, tol=1e-16):
+        """
+        Check recently computed totals derivative jacobian for problems.
+
+        Some optimizers can't handle a jacobian when a design variable has no effect on the
+        constraints, or a constraint is unaffected by a design variable. This method
+        checks for these cases.
+
+        Parameters
+        ----------
+        tol : double
+            Tolerance for the check.
+        raise_error : bool
+            If True, raise an exception if a zero row or column is found.
+        """
+        J = np.abs(self.J)
+        nrows, ncols = J.shape
+        zero_rows = []
+        zero_cols = []
+
+        # Check for zero rows, which correspond to constraints unaffected by any design vars.
+        for j in np.arange(nrows):
+            if np.all(J[j, :] < tol):
+                for name, val in self.of_meta.items():
+                    if j > val[0].stop - 1:
+                        continue
+                    break
+                if name in self.ivc_print_names:
+                    name = self.ivc_print_names[name]
+                if name not in zero_rows:
+                    zero_rows.append(name)
+
+        if zero_rows:
+            msg = f"Constraints or objectives {zero_rows} cannot be impacted by the design " + \
+                "variables of the problem."
+            if raise_error:
+                raise RuntimeError(msg)
+            else:
+                issue_warning(msg, category=DerivativesWarning)
+
+        # Check for zero cols, which correspond to design vars that don't affect anything.
+        for j in np.arange(ncols):
+            if np.all(J[:, j] < tol):
+                for name, val in self.wrt_meta.items():
+                    if j > val[0].stop - 1:
+                        continue
+                    break
+                if name in self.ivc_print_names:
+                    name = self.ivc_print_names[name]
+                if name not in zero_cols:
+                    zero_cols.append(name)
+
+        if zero_cols:
+            msg = f"Design variables {zero_cols} have no impact on the constraints or objective."
+            if raise_error:
+                raise RuntimeError(msg)
+            else:
+                issue_warning(msg, category=DerivativesWarning)
+
     def _restore_linear_solution(self, vec_names, key, mode):
         """
         Restore the previous linear solution.
@@ -1765,6 +1832,8 @@ class _TotalJacInfo(object):
             J_dict = self.J_dict
             for of, wrt_dict in J_dict.items():
                 for wrt, J_sub in wrt_dict.items():
+                    if wrt in self.ivc_print_names:
+                        wrt = self.ivc_print_names[wrt]
                     pprint.pprint({(of, wrt): J_sub})
         else:
             J = self.J
@@ -1774,7 +1843,10 @@ class _TotalJacInfo(object):
                 out_slice = self.of_meta[of][0]
                 for j, wrt in enumerate(self.wrt):
                     if wrt not in self.remote_vois:
-                        pprint.pprint({(of, wrt): J[out_slice, self.wrt_meta[wrt][0]]})
+                        deriv = J[out_slice, self.wrt_meta[wrt][0]]
+                        if wrt in self.ivc_print_names:
+                            wrt = self.ivc_print_names[wrt]
+                        pprint.pprint({(of, wrt): deriv})
 
         print('')
         sys.stdout.flush()
@@ -1801,7 +1873,7 @@ class _TotalJacInfo(object):
             self.model._recording_iter.pop()
 
 
-def _get_subjac(jac_meta, prom_out, prom_in, of_idx, wrt_idx, dist_resp, comm):
+def _get_approx_subjac(jac_meta, prom_out, prom_in, of_idx, wrt_idx, dist_resp, comm):
     """
     Return proper subjacobian based on input/output names and indices.
 
@@ -1813,9 +1885,9 @@ def _get_subjac(jac_meta, prom_out, prom_in, of_idx, wrt_idx, dist_resp, comm):
         Promoted output name.
     prom_in : str
         Promoted input name.
-    of_idx : ndarray or None
+    of_idx : ndarray or slice
         Output indices, if any.
-    wrt_idx : ndarray or None
+    wrt_idx : ndarray or slice
         Input indices, if any.
     dist_resp : None or tuple
         Tuple containing indices and sizes if this response is distributed.
@@ -1827,13 +1899,23 @@ def _get_subjac(jac_meta, prom_out, prom_in, of_idx, wrt_idx, dist_resp, comm):
     ndarray
         The desired subjacobian.
     """
-    if jac_meta['rows'] is not None:  # sparse list format
-        # This is a design variable that was declared as an obj/con.
-        tot = np.eye(len(jac_meta['value']))
-        if of_idx is not None:
-            tot = tot[of_idx, :]
-        if wrt_idx is not None:
-            tot = tot[:, wrt_idx]
+    # of_idx and wrt_idx are relative to the full sized of and wrt vars, but
+    # 'rows' and 'cols' are relative to the subjac after compression based on of_idx
+    # and wrt_idx.
+    if jac_meta['rows'] is not None:
+        if prom_in == prom_out:
+            nrows, ncols = jac_meta['shape']
+            if of_idx is _full_slice or wrt_idx is _full_slice:
+                sz = nrows  # nrows = ncols since this subjac is square
+            else:
+                sz = np.max(of_idx)
+                sz = max(sz, np.max(wrt_idx)) + 1
+            # take sections of the identity matrix based on of_idx and wrt_idx
+            sz = max(sz, nrows, ncols)
+            tot = np.eye(sz)[of_idx, wrt_idx]
+        else:
+            tot = np.zeros(jac_meta['shape'])
+            tot[jac_meta['rows'], jac_meta['cols']] = jac_meta['value']
     else:
         tot = jac_meta['value']
 
