@@ -9,8 +9,6 @@ additional MPI capability.
 from collections import OrderedDict
 import json
 import signal
-import sys
-import traceback
 
 import numpy as np
 from scipy.sparse import coo_matrix
@@ -20,24 +18,25 @@ try:
 except ImportError:
     Optimization = None
 
+from openmdao.core.constants import INT_DTYPE
 from openmdao.core.analysis_error import AnalysisError
 from openmdao.core.driver import Driver, RecordingDebugging
 import openmdao.utils.coloring as c_mod
-from openmdao.utils.general_utils import simple_warning
 from openmdao.utils.class_util import weak_method_wrapper
 from openmdao.utils.mpi import FakeComm
+from openmdao.warnings import issue_warning, DerivativesWarning
 
 
 # names of optimizers that use gradients
 grad_drivers = {'CONMIN', 'FSQP', 'IPOPT', 'NLPQLP',
-                'PSQP', 'SLSQP', 'SNOPT', 'NLPY_AUGLAG'}
+                'PSQP', 'SLSQP', 'SNOPT', 'NLPY_AUGLAG', 'ParOpt'}
 
 # names of optimizers that allow multiple objectives
 multi_obj_drivers = {'NSGA2'}
 
 # All optimizers in pyoptsparse
 optlist = ['ALPSO', 'CONMIN', 'FSQP', 'IPOPT', 'NLPQLP',
-           'NSGA2', 'PSQP', 'SLSQP', 'SNOPT', 'NLPY_AUGLAG', 'NOMAD']
+           'NSGA2', 'PSQP', 'SLSQP', 'SNOPT', 'NLPY_AUGLAG', 'NOMAD', 'ParOpt']
 
 # All optimizers that require an initial run
 run_required = ['NSGA2', 'ALPSO']
@@ -49,23 +48,40 @@ DEFAULT_OPT_SETTINGS['IPOPT'] = {
     'linear_solver': 'mumps'
 }
 
-CITATIONS = """@article{Hwang_maud_2018
- author = {Hwang, John T. and Martins, Joaquim R.R.A.},
- title = "{A Computational Architecture for Coupling Heterogeneous
-          Numerical Models and Computing Coupled Derivatives}",
- journal = "{ACM Trans. Math. Softw.}",
- volume = {44},
- number = {4},
- month = jun,
- year = {2018},
- pages = {37:1--37:39},
- articleno = {37},
- numpages = {39},
- doi = {10.1145/3182393},
- publisher = {ACM},
+CITATIONS = """@article{Wu_pyoptsparse_2020,
+    author = {Neil Wu and Gaetan Kenway and Charles A. Mader and John Jasa and
+     Joaquim R. R. A. Martins},
+    title = {{pyOptSparse:} A {Python} framework for large-scale constrained
+     nonlinear optimization of sparse systems},
+    journal = {Journal of Open Source Software},
+    volume = {5},
+    number = {54},
+    month = {October},
+    year = {2020},
+    pages = {2564},
+    doi = {10.21105/joss.02564},
+    publisher = {The Open Journal},
+}
+
+@article{Hwang_maud_2018
+    author = {Hwang, John T. and Martins, Joaquim R.R.A.},
+    title = "{A Computational Architecture for Coupling Heterogeneous
+             Numerical Models and Computing Coupled Derivatives}",
+    journal = "{ACM Trans. Math. Softw.}",
+    volume = {44},
+    number = {4},
+    month = jun,
+    year = {2018},
+    pages = {37:1--37:39},
+    articleno = {37},
+    numpages = {39},
+    doi = {10.1145/3182393},
+    publisher = {ACM},
 }
 """
 
+DEFAULT_SIGNAL = None
+
 
 class UserRequestedException(Exception):
     """
@@ -87,7 +103,7 @@ class pyOptSparseDriver(Driver):
     constrained optimization problems, with additional MPI capability.
     pypptsparse has interfaces to the following optimizers:
     ALPSO, CONMIN, FSQP, IPOPT, NLPQLP, NSGA2, PSQP, SLSQP,
-    SNOPT, NLPY_AUGLAG, NOMAD.
+    SNOPT, NLPY_AUGLAG, NOMAD, ParOpt.
     Note that some of these are not open source and therefore not included
     in the pyoptsparse source code.
 
@@ -111,6 +127,10 @@ class pyOptSparseDriver(Driver):
         Dictionary for setting optimizer-specific options.
     pyopt_solution : Solution
         Pyopt_sparse solution object.
+    _check_jac : bool
+        Used internally to control when to perform singular checks on computed total derivs.
+    _exc_info : None or <Exception>
+        Cached exception that was raised in the _objfunc or _gradfunc callbacks.
     _in_user_function :bool
         This is set to True at the start of a pyoptsparse callback to _objfunc and _gradfunc, and
         restored to False at the finish of each callback.
@@ -120,7 +140,7 @@ class pyOptSparseDriver(Driver):
         Contains the objectives plus nonlinear constraints.
     _signal_cache : <Function>
         Cached function pointer that was assigned as handler for signal defined in option
-        user_teriminate_signal.
+        user_terminate_signal.
     _user_termination_flag : bool
         This is set to True when the user sends a signal to terminate the job.
     """
@@ -137,7 +157,7 @@ class pyOptSparseDriver(Driver):
         if Optimization is None:
             raise RuntimeError('pyOptSparseDriver is not available, pyOptsparse is not installed.')
 
-        super(pyOptSparseDriver, self).__init__(**kwargs)
+        super().__init__(**kwargs)
 
         # What we support
         self.supports['inequality_constraints'] = True
@@ -151,6 +171,7 @@ class pyOptSparseDriver(Driver):
         # What we don't support yet
         self.supports['active_set'] = False
         self.supports['integer_design_vars'] = False
+        self.supports['distributed_design_vars'] = False
         self.supports._read_only = True
 
         # The user places optimizer-specific settings in here.
@@ -172,6 +193,8 @@ class pyOptSparseDriver(Driver):
         self._signal_cache = None
         self._user_termination_flag = False
         self._in_user_function = False
+        self._check_jac = False
+        self._exc_info = None
 
         self.cite = CITATIONS
 
@@ -188,9 +211,25 @@ class pyOptSparseDriver(Driver):
         self.options.declare('gradient method', default='openmdao',
                              values={'openmdao', 'pyopt_fd', 'snopt_fd'},
                              desc='Finite difference implementation to use')
-        self.options.declare('user_teriminate_signal', default=signal.SIGUSR1, allow_none=True,
-                             desc='OS signal that triggers a clean user-terimnation. Only SNOPT'
+        self.options.declare('user_terminate_signal', default=DEFAULT_SIGNAL, allow_none=True,
+                             desc='OS signal that triggers a clean user-termination. Only SNOPT'
                              'supports this option.')
+        self.options.declare('singular_jac_behavior', default='warn',
+                             values=['error', 'warn', 'ignore'],
+                             desc='Defines behavior of a zero row/col check after first call to'
+                             'compute_totals:'
+                             'error - raise an error.'
+                             'warn - raise a warning.'
+                             "ignore - don't perform check.")
+        self.options.declare('singular_jac_tol', default=1e-16,
+                             desc='Tolerance for zero row/column check.')
+
+        # Deprecated option
+        self.options.declare('user_teriminate_signal', default=None, allow_none=True,
+                             desc='OS signal that triggers a clean user-termination. Only SNOPT'
+                             'supports this option.',
+                             deprecation="The option 'user_teriminate_signal' was misspelled and "
+                             "will be deprecated. Please use 'user_terminate_signal' instead.")
 
     def _setup_driver(self, problem):
         """
@@ -203,7 +242,7 @@ class pyOptSparseDriver(Driver):
         problem : <Problem>
             Pointer to the containing problem.
         """
-        super(pyOptSparseDriver, self)._setup_driver(problem)
+        super()._setup_driver(problem)
 
         self.supports._read_only = False
         self.supports['gradients'] = self.options['optimizer'] in grad_drivers
@@ -216,6 +255,11 @@ class pyOptSparseDriver(Driver):
 
         self._setup_tot_jac_sparsity()
 
+        # Handle deprecated option.
+        if self.options._dict['user_teriminate_signal']['value'] is not None:
+            self.options['user_terminate_signal'] = \
+                self.options._dict['user_teriminate_signal']['value']
+
     def run(self):
         """
         Excute pyOptsparse.
@@ -239,6 +283,7 @@ class pyOptSparseDriver(Driver):
         self._quantities = []
 
         self._check_for_missing_objective()
+        self._check_jac = self.options['singular_jac_behavior'] in ['error', 'warn']
 
         # Only need initial run if we have linear constraints or if we are using an optimizer that
         # doesn't perform one initially.
@@ -266,22 +311,23 @@ class pyOptSparseDriver(Driver):
                 info = self._coloring_info
                 if info['min_improve_pct'] > pct:
                     info['coloring'] = info['static'] = None
-                    simple_warning("%s: Coloring was deactivated.  Improvement of %.1f%% was less "
-                                   "than min allowed (%.1f%%)." % (self.msginfo, pct,
-                                                                   info['min_improve_pct']))
+                    msg = f"Coloring was deactivated.  Improvement of {pct:.1f}% was less " \
+                          f"than min allowed ({info['min_improve_pct']:.1f}%)."
+                    issue_warning(msg, prefix=self.msginfo, category=DerivativesWarning)
 
         comm = None if isinstance(problem.comm, FakeComm) else problem.comm
         opt_prob = Optimization(self.options['title'], weak_method_wrapper(self, '_objfunc'),
                                 comm=comm)
 
         # Add all design variables
-        param_meta = self._designvars
-        self._indep_list = indep_list = list(param_meta)
-        param_vals = self.get_design_var_values()
+        input_meta = self._designvars
+        self._indep_list = indep_list = list(input_meta)
+        input_vals = self.get_design_var_values()
 
-        for name, meta in param_meta.items():
-            opt_prob.addVarGroup(name, meta['size'], type='c',
-                                 value=param_vals[name],
+        for name, meta in input_meta.items():
+            size = meta['global_size'] if meta['distributed'] else meta['size']
+            opt_prob.addVarGroup(name, size, type='c',
+                                 value=input_vals[name],
                                  lower=meta['lower'], upper=meta['upper'])
 
         opt_prob.finalizeDesignVariables()
@@ -319,10 +365,10 @@ class pyOptSparseDriver(Driver):
             size = meta['global_size'] if meta['distributed'] else meta['size']
             lower = upper = meta['equals']
             if fwd:
-                wrt = [v for v in indep_list if name in relevant[v]]
+                wrt = [v for v in indep_list if name in relevant[input_meta[v]['ivc_source']]]
             else:
                 rels = relevant[name]
-                wrt = [v for v in indep_list if v in rels]
+                wrt = [v for v in indep_list if input_meta[v]['ivc_source'] in rels]
 
             if meta['linear']:
                 jac = {w: _lin_jacs[name][w] for w in wrt}
@@ -331,7 +377,7 @@ class pyOptSparseDriver(Driver):
             else:
                 if name in self._res_jacs:
                     resjac = self._res_jacs[name]
-                    jac = {n: resjac[n] for n in wrt}
+                    jac = {n: resjac[input_meta[n]['ivc_source']] for n in wrt}
                 else:
                     jac = None
                 opt_prob.addConGroup(name, size, lower=lower, upper=upper, wrt=wrt, jac=jac)
@@ -348,10 +394,10 @@ class pyOptSparseDriver(Driver):
             upper = meta['upper']
 
             if fwd:
-                wrt = [v for v in indep_list if name in relevant[v]]
+                wrt = [v for v in indep_list if name in relevant[input_meta[v]['ivc_source']]]
             else:
                 rels = relevant[name]
-                wrt = [v for v in indep_list if v in rels]
+                wrt = [v for v in indep_list if input_meta[v]['ivc_source'] in rels]
 
             if meta['linear']:
                 jac = {w: _lin_jacs[name][w] for w in wrt}
@@ -360,7 +406,7 @@ class pyOptSparseDriver(Driver):
             else:
                 if name in self._res_jacs:
                     resjac = self._res_jacs[name]
-                    jac = {n: resjac[n] for n in wrt}
+                    jac = {n: resjac[input_meta[n]['ivc_source']] for n in wrt}
                 else:
                     jac = None
                 opt_prob.addConGroup(name, size, upper=upper, lower=lower, wrt=wrt, jac=jac)
@@ -387,33 +433,44 @@ class pyOptSparseDriver(Driver):
         for option, value in self.opt_settings.items():
             opt.setOption(option, value)
 
-        # Execute the optimization problem
-        if self.options['gradient method'] == 'pyopt_fd':
+        self._exc_info = None
+        try:
 
-            # Use pyOpt's internal finite difference
-            # TODO: Need to get this from OpenMDAO
-            # fd_step = problem.model.deriv_options['step_size']
-            fd_step = 1e-6
-            sol = opt(opt_prob, sens='FD', sensStep=fd_step, storeHistory=self.hist_file,
-                      hotStart=self.hotstart_file)
+            # Execute the optimization problem
+            if self.options['gradient method'] == 'pyopt_fd':
 
-        elif self.options['gradient method'] == 'snopt_fd':
-            if self.options['optimizer'] == 'SNOPT':
-
-                # Use SNOPT's internal finite difference
+                # Use pyOpt's internal finite difference
                 # TODO: Need to get this from OpenMDAO
                 # fd_step = problem.model.deriv_options['step_size']
                 fd_step = 1e-6
-                sol = opt(opt_prob, sens=None, sensStep=fd_step, storeHistory=self.hist_file,
+                sol = opt(opt_prob, sens='FD', sensStep=fd_step, storeHistory=self.hist_file,
                           hotStart=self.hotstart_file)
 
-            else:
-                raise Exception("SNOPT's internal finite difference can only be used with SNOPT")
-        else:
+            elif self.options['gradient method'] == 'snopt_fd':
+                if self.options['optimizer'] == 'SNOPT':
 
-            # Use OpenMDAO's differentiator for the gradient
-            sol = opt(opt_prob, sens=weak_method_wrapper(self, '_gradfunc'),
-                      storeHistory=self.hist_file, hotStart=self.hotstart_file)
+                    # Use SNOPT's internal finite difference
+                    # TODO: Need to get this from OpenMDAO
+                    # fd_step = problem.model.deriv_options['step_size']
+                    fd_step = 1e-6
+                    sol = opt(opt_prob, sens=None, sensStep=fd_step, storeHistory=self.hist_file,
+                              hotStart=self.hotstart_file)
+
+                else:
+                    msg = "SNOPT's internal finite difference can only be used with SNOPT"
+                    self._exc_info = Exception(msg)
+            else:
+
+                # Use OpenMDAO's differentiator for the gradient
+                sol = opt(opt_prob, sens=weak_method_wrapper(self, '_gradfunc'),
+                          storeHistory=self.hist_file, hotStart=self.hotstart_file)
+
+        except Exception as _:
+            if not self._exc_info:
+                raise()
+
+        if self._exc_info:
+            raise self._exc_info
 
         # Print results
         if self.options['print_results']:
@@ -426,7 +483,11 @@ class pyOptSparseDriver(Driver):
             self.set_design_var(name, dv_dict[name])
 
         with RecordingDebugging(self._get_name(), self.iter_count, self) as rec:
-            model.run_solve_nonlinear()
+            try:
+                model.run_solve_nonlinear()
+            except AnalysisError:
+                model._clear_iprint()
+
             rec.abs = 0.0
             rec.rel = 0.0
         self.iter_count += 1
@@ -450,7 +511,7 @@ class pyOptSparseDriver(Driver):
             pass
 
         # revert signal handler to cached version
-        sigusr = self.options['user_teriminate_signal']
+        sigusr = self.options['user_terminate_signal']
         if sigusr is not None:
             signal.signal(sigusr, self._signal_cache)
             self._signal_cache = None   # to prevent memory leak test from failing
@@ -483,7 +544,7 @@ class pyOptSparseDriver(Driver):
 
         # Note: we place our handler as late as possible so that codes that run in the
         # workflow can place their own handlers.
-        sigusr = self.options['user_teriminate_signal']
+        sigusr = self.options['user_terminate_signal']
         if sigusr is not None and self._signal_cache is None:
             self._signal_cache = signal.getsignal(sigusr)
             signal.signal(sigusr, self._signal_handler)
@@ -526,13 +587,8 @@ class pyOptSparseDriver(Driver):
                 rec.abs = 0.0
                 rec.rel = 0.0
 
-        except Exception as msg:
-            tb = traceback.format_exc()
-
-            # Exceptions seem to be swallowed by the C code, so this
-            # should give the user more info than the dreaded "segfault"
-            print("Exception: %s" % str(msg))
-            print(70 * "=", tb, 70 * "=")
+        except Exception as raised:
+            self._exc_info = raised
             fail = 1
             func_dict = {}
 
@@ -580,6 +636,14 @@ class pyOptSparseDriver(Driver):
                 sens_dict = self._compute_totals(of=self._quantities,
                                                  wrt=self._indep_list,
                                                  return_format='dict')
+
+                # First time through, check for zero row/col.
+                if self._check_jac:
+                    raise_error = self.options['singular_jac_behavior'] == 'error'
+                    self._total_jac.check_total_jac(raise_error=raise_error,
+                                                    tol=self.options['singular_jac_tol'])
+                    self._check_jac = False
+
             # Let the optimizer try to handle the error
             except AnalysisError:
                 prob.model._clear_iprint()
@@ -598,10 +662,12 @@ class pyOptSparseDriver(Driver):
                 res_jacs = self._res_jacs
                 for okey in func_dict:
                     new_sens[okey] = newdv = OrderedDict()
+                    okey_src = self._responses[okey]['ivc_source']
                     for ikey in dv_dict:
-                        if okey in res_jacs and ikey in res_jacs[okey]:
+                        ikey_src = self._designvars[ikey]['ivc_source']
+                        if okey_src in res_jacs and ikey_src in res_jacs[okey_src]:
                             arr = sens_dict[okey][ikey]
-                            coo = res_jacs[okey][ikey]
+                            coo = res_jacs[okey_src][ikey_src]
                             row, col, data = coo['coo']
                             coo['coo'][2] = arr[row, col].flatten()
                             newdv[ikey] = coo
@@ -621,13 +687,9 @@ class pyOptSparseDriver(Driver):
                         isize = len(ival)
                         sens_dict[okey][ikey] = np.zeros((osize, isize))
 
-        except Exception as msg:
-            tb = traceback.format_exc()
-
-            # Exceptions seem to be swallowed by the C code, so this
-            # should give the user more info than the dreaded "segfault"
-            print("Exception: %s" % str(msg))
-            print(70 * "=", tb, 70 * "=", flush=True)
+        except Exception as raised:
+            self._exc_info = raised
+            fail = 1
             sens_dict = {}
 
         # print("Derivatives calculated")
@@ -703,8 +765,8 @@ class pyOptSparseDriver(Driver):
                 continue
             self._res_jacs[res] = {}
             for dv, (rows, cols, shape) in resdict.items():
-                rows = np.array(rows, dtype=int)
-                cols = np.array(cols, dtype=int)
+                rows = np.array(rows, dtype=INT_DTYPE)
+                cols = np.array(cols, dtype=INT_DTYPE)
 
                 self._res_jacs[res][dv] = {
                     'coo': [rows, cols, np.zeros(rows.size)],
